/*
    This file is part of COMMON AVR UTILS.

    COMMON AVR UTILS is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    COMMON AVR UTILS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

#include <stdarg.h>

#include "uput.h"
#include "div.h"
#include "pgmread-inc.h"

#define PUT_DEC2           (1 <<  0)
#define PUT_UINT16         (1 <<  1)
#define PUT_SINT16         (1 <<  2)
#define PUT_HEX2           (1 <<  3)

#define PUT_HEX4           (1 <<  4)
#define PUT_HEX8           (1 <<  5)
#define PUT_BIN8           (1 <<  6)
#define PUT_SFRAC16        (1 <<  7)

#define PUT_SFRAC32        (1 <<  8)
#define PUT_SFRAC12        (1 <<  9)
#define PUT_FUNC           (1 << 10)
#define PUT_AT             (1 << 11)

#define PUT_MEM            (1 << 12)
#define PUT_SINT32         (1 << 13)

#ifndef PUTTER
#define PUTTER (-1)
#endif

#ifdef PUT_CHAR_INIT
extern void PUT_CHAR_INIT (char);
#define put_dummy PUT_CHAR_INIT
#else
static void put_dummy (char c) { (void) c;}
#endif // PUT_CHAR_INIT

void (*put_char)(char) = put_dummy;

// Char output function, defaults to <put_char> and may be changed by %F
static void (*put_local)(char);

// Introduce put_char2 to avoid indirect calls below
static void __attribute__ ((noinline)) put_char2     (char);
void put_char2 (char c)
{
    put_local (c);
}

// Level 1
#define NEED_AT      ((PUTTER) & PUT_AT)
#define NEED_FUNC    ((PUTTER) & PUT_FUNC)
#define NEED_MEM     ((PUTTER) & PUT_MEM)
#define NEED_DEC2    ((PUTTER) & PUT_DEC2)
#define NEED_HEX8    ((PUTTER) & PUT_HEX8)
#define NEED_BIN8    ((PUTTER) & PUT_BIN8)
#define NEED_SINT16  ((PUTTER) & PUT_SINT16)
#define NEED_SINT32  ((PUTTER) & PUT_SINT32)
#define NEED_SFRAC12 ((PUTTER) & PUT_SFRAC12)
#define NEED_SFRAC16 ((PUTTER) & PUT_SFRAC16)
#define NEED_SFRAC32 ((PUTTER) & PUT_SFRAC32)

// Level 2
#define NEED_UINT16  (((PUTTER) & PUT_UINT16)  || NEED_SINT16 || NEED_MEM || NEED_DEC2)
#define NEED_HEX4    (((PUTTER) & PUT_HEX4)    || NEED_HEX8)

#define NEED_SFRAC32_FRAC (NEED_SFRAC16 || NEED_SFRAC12 || NEED_SFRAC32)
#define NEED_SFRAC16_INT  (NEED_SFRAC16 || NEED_SFRAC12)
#define NEED_MAC10        (NEED_SINT32 || NEED_SFRAC32)

// Level 3
#define NEED_HEX2    (((PUTTER) & PUT_HEX2) || NEED_HEX4)

#define HAVE_32BIT_FUNCS (NEED_SINT32 || NEED_SFRAC32 || NEED_HEX8)

#define NEED_BUF16   (NEED_BIN8 || PUT_UINT16)

#if NEED_BUF16
static char buf[16];
#endif

// I am using this code throughout my projects, so I try
// to get the best out of gcc. Sometimes it performs some
// silly "optimisations" because it known some value at
// compile time.
// HIDE hides this value from gcc. It does not generate any
// by itself, but it infuences code generation of gcc
#define HIDE(VAL) __asm ("; HIDE " #VAL : "+r" (VAL))

// gcc 4.3.2 will throw an ICE otherways
#if __GNUC__ >= 4
#define asm(...)
#endif

#if NEED_AT
// %@: write as 0-terminated string to address <addr>
static void put_at (char *);
static void put_addr (char);
static char * addr;

// %@: write as 0-terminated string to address <addr>
void put_at (char * a)
{
    *a = '\0';
    addr = a;
    put_local = put_addr;
}

// Write <c> to the address set in put_at
void put_addr (char c)
{
    *addr++ = c;
    *addr   = 0;
}
#endif // NEED_AT

#if NEED_FUNC
// %F: set char output function to <func>
static void put_func (void (*func)(char));
void put_func (void (*func)(char))
{
    put_local = func;
}
#endif // NEED_FUNC

// %s: Put 0-terminated string from RAM
void put_str (const char *s)
{
    while (1)
    {
        char c = *s++;
        if ('\0' == c)
            break;
        put_char2 (c);
    }
}

// %S: Put 0-terminated string from FLASH
void put_pstr (PGM_P s)
{
    while (1)
    {
        char c = pgm_read_byte (s);
        s++;
        if ('\0' == c)
            break;
        put_char2 (c);
    }
}

#if NEED_UINT16
// %u: Put unsigned 16-bit integer, no leading zeros
static void put_uint16  (uint16_t);
void put_uint16 (uint16_t zahl)
{
    // 5 Ziffern (0...65535) und evtl. noch eine fhrende 0
    // +1 wegen String-Ende (wird im Startup auf 0 gesetzt)

    // s zeigt auf das Ende von string
    // string wird von hinten nach vorne gefllt
    char *s = buf + 5;

    *s = '\0';

    // qrem enthlt Quotient (quot) und Rest (rem) der Divisionen
    udiv_t qrem = {.quot = zahl};

    do
    {
        // Division mit Rest durch 10
        // quot: Ergebnis fr den nchsten Durchlauf
        // rem:  Rest ist die Ziffer im 10er-System
        qrem = udiv (qrem.quot, 10);

        // Ziffer in Zeichen wandeln und speichern
        *(--s) = '0' + qrem.rem;
    }
    while (0 != qrem.quot);

    // Falls eine fhrende '0' gespeichert wurde: weg damit
    // ausser Zahl war selbst schon 0
    if (*s == '0' && *(s+1) != '\0')
        s++;

    put_str (s);
}
#endif

#if NEED_SINT16
// %d: Put signed 16-bit integer, no leading zeros
static void put_sint16   (int16_t);
void put_sint16 (int16_t zahl)
{
    if (zahl < 0)
    {
        put_char2 ('-');
        zahl = -zahl;
    }

    put_uint16 ((uint16_t) zahl);
}
#endif

#if NEED_MEM
#include "mem-check.h"
static void put_mem (void)
{
    put_uint16 (get_mem_unused());
}
#endif // PUT_MEM

#if NEED_DEC2
// %2: Put 8-Bit integer als decimal number with exact 2 decimal digits
static void put_dec2    (uint8_t);
void put_dec2 (uint8_t i)
{
    if (i > 99)
    {
        put_char2 ('-');
        put_char2 ('-');
        return;
    }

    if (i < 10)
        put_char2 ('0');

    put_uint16 (i);
}
#endif

#if NEED_HEX2
// %x: Put 2 nibbles of 8-Bit intger as hex digits
static void put_nibble (uint8_t);
void put_nibble (uint8_t nibble)
{
    nibble += '0';
    if (nibble >= '0'+10)
        nibble += 'a'-'0'-10;

    put_char2 (nibble);
}

static void put_hex2 (uint8_t);
void put_hex2 (uint8_t c)
{
    // hi nibble
    put_nibble (c >> 4);

    // lo nibble
    put_nibble (c & 0x0f);
}
#endif // PUT_HEX2 || PUT_HEX4

#if NEED_HEX4
// %X: Put 16-Bit intger as hex like ****
static void put_hex4 (uint16_t);
void put_hex4 (uint16_t c)
{
    // hi byte
    put_hex2 (c >> 8);

    // lo byte
    put_hex2 (c);
}
#endif // PUT_HEX4

#if NEED_HEX8
// %Y: Put 32-Bit intger as hex like ********
static void put_hex8 (uint32_t);
void put_hex8 (uint32_t c)
{
    // hi word
    put_hex4 (c >> 16);

    // lo byte
    put_hex4 (c);
}
#endif // PUT_HEX8

#if NEED_BIN8
// %b: Put 8-bit number as binary with exactly 8 digits
static void put_bin8  (uint8_t);
void put_bin8 (uint8_t c)
{
    uint8_t ndigits = 8;
    char * pbuf = buf+8;

    buf[8] = '\0';

    do
    {
        char dig = '0';

        if (c & 1)
            dig = '1';

        *--pbuf = dig;

        c >>= 1;
    } while (--ndigits);

    put_str (buf);
}
#endif // PUT_BIN8

#if NEED_SFRAC32 || NEED_SFRAC16 || NEED_SINT32
// %q, %Q, %v, %l
// Helper for long and signed q-Format

#include "data-types.h"

#define MAC10_STEP(D16,Q,N)         \
do {                                \
      D16.asByte[0] = Q.asByte[N];  \
      D16 = mac10 (D16);            \
      Q.asByte[N] = D16.asByte[0];  \
} while(0)
      
static data16_t __attribute__((noinline)) mac10 (data16_t data);

data16_t mac10 (data16_t data)
{
    data.asWord = data.asByte[0] * 10 + data.asByte[1];
    return data;
}
#endif

#if NEED_SINT32
static void __attribute__((noinline)) put_sint32  (data32_t);

void put_sint32 (data32_t q)
{
    uint8_t is0 = 1;
    
    if (q.asLong < 0)
    {
        put_char2 ('-');
        q.asLong = -q.asLong;
    }
    
    for (uint8_t i=0;; i++)
    {
        int8_t ziffer = 0;
        
        uint32_t E8 = 1E8;
        HIDE (E8);
        
        if (q.asByte[3] >= 100000000ul/0x1000000ul) 
        {
            do
            {
                ziffer++;
                q.asLong -= E8;
            }
            while (q.asLong >= 0);
            
            q.asLong += E8;
            ziffer--;
            
            uint8_t ziffer2 = '0'-1;
            
            do
            {
                ziffer2++;
                ziffer -= 10;
            }
            while (ziffer >= 0);
            
            ziffer += 10;
            
            if (ziffer2 != '0')
            {
                put_char2 (ziffer2);
                is0 = 0;
            }
                
            if (ziffer != '0')
                is0 = 0;
        }

        if (!is0 || i == 8)
        {
            put_char2 (ziffer+'0');
            if (i == 8)
                break;
        }
        
        register data16_t digit asm ("r24");
        digit.asByte[1] = 0; // carry = 0
        
        MAC10_STEP (digit, q, 0);
        MAC10_STEP (digit, q, 1);
        MAC10_STEP (digit, q, 2);
        MAC10_STEP (digit, q, 3);
    }
}
#endif // SINT32

#if NEED_SFRAC32_FRAC
// Helper for Q-format %q, %Q, %v:
// Print just <nigs> digits the positive fractional part
void __attribute__((noinline)) put_sfrac32_frac (data32_t, uint8_t);
void put_sfrac32_frac (data32_t q, uint8_t ndigs)
{
    put_char2 ('.');
    
    for (uint8_t i=0; i < ndigs; i++)
    {
        register data16_t digit asm ("r24");
        digit.asByte[1] = 0; // carry = 0
        
        MAC10_STEP (digit, q, 0);
        MAC10_STEP (digit, q, 1);
        MAC10_STEP (digit, q, 2);
        MAC10_STEP (digit, q, 3);
        
        put_char2 ('0' + digit.asByte[1]);
    }
}
#endif // SFRAC32_DOT

#if NEED_SFRAC16_INT
// Helper for Q-format %q, %v:
// Print just <nigs> digits of integral part
// and return fractional part as 1.31 Q-format
data32_t __attribute__((noinline)) put_sfrac16_int (data16_t, uint8_t);
data32_t put_sfrac16_int (data16_t q, uint8_t shift)
{
    if (q.asShort < 0)
    {
        put_char2 ('-');
        q.asShort = -q.asShort;
    }
    
    data32_t q32 = {.asULong = q.asWord };
    
    do
    {
        q32.asULong <<= 1;
    } while (--shift);
    
    put_uint16 (q32.asWord[1]);
    
    q32.asULong <<= 16;
    
    return q32;
}
#endif // SFRAC16_INT

#if NEED_SFRAC16
// %q: Put signed 1.15 q-format fractional as decimal
static void __attribute__((noinline)) put_sfrac16  (data16_t);

void put_sfrac16 (data16_t q)
{
    put_sfrac32_frac (put_sfrac16_int (q, 1), 5);
}
#endif // PUT_SFRAC16

#if NEED_SFRAC12
// %v: Put signed 4.12 q-format fractional as decimal
static void put_sfrac12  (data16_t);
void put_sfrac12 (data16_t q)
{
    put_sfrac32_frac (put_sfrac16_int (q, 4), 4);
}
#endif // SFRAC12

#if NEED_SFRAC32
// %Q: Put signed 1.31 q-format fractional as decimal
void put_sfrac32 (data32_t q)
{
    if (q.asLong < 0)
    {
        put_char2 ('-');
        q.asLong = -q.asLong;
    }

    char digit0 = '0';
    HIDE (digit0);
    
    if ((q.asByte[3] & 0x80))
        digit0 = '1';
    put_char2 (digit0);
    
    q.asULong <<= 1;
    
    put_sfrac32_frac (q, 10);
}
#endif // PUT_SFRAC32

/*
Strings, Chars
    %s :  String (RAM)
    %S :  String (Flash)
    %c :  one char
    %% :  %

Decimal
    %2 :  unsigned  8-bit dec-val, 2 digits
    %u :  unsigned 16-bit int, no leading zeros
    %d :  signed   16-bit int, no leading zeros
    %l :  signed   32-bit int, no leading zeros

Hex
    %x :  unsigned  8-bit hex-val, 2 digits
    %X :  unsigned 16-bit hex-val, 4 digits
    %Y :  unsigned 32-bit hex-val, 8 digits

Binary
    %b :  unsigned  8-bit bin-val, 8 digits

Q-Format
    %q :  signed 1.15-bit fraction
    %Q :  signed 1.31-bit fraction
    %v :  signed 4.12-bit fraction

Misc
    %F :  set output function void (*)(char)
    %@ :  write to respective address
          resulting string will be 0-terminated
*/

#if HAVE_32BIT_FUNCS
typedef uint32_t putter_arg_t;
#else
typedef uint16_t putter_arg_t;
#endif // HAVE_32BIT_FUNCS

typedef void (*putter_t)(putter_arg_t);

static const struct putters
{
    char id;
    void * func;
} fn_map[] PROGMEM =
{
//Functions that require a 16-Bit argument.
// Last must be 'c'.
#if NEED_UINT16
    { 'u', put_uint16 },
#endif
#if NEED_SINT16
    { 'd', put_sint16 },
#endif
#if NEED_HEX2
    { 'x', put_hex2   },
#endif
#if NEED_HEX4
    { 'X', put_hex4   },
#endif
#if NEED_DEC2
    { '2', put_dec2   },
#endif
#if NEED_BIN8
    { 'b', put_bin8   },
#endif
#if NEED_AT
    { '@', put_at     },
#endif
#if NEED_FUNC
    { 'F', put_func   },
#endif
#if NEED_SFRAC16
    { 'q', put_sfrac16 },
#endif
#if NEED_SFRAC12
    { 'v', put_sfrac12 },
#endif
    { 's', put_str    },
    { 'S', put_pstr   },
    { 'c', put_char2  }
    
//Functions that require a 32-Bit argument follow 'c':

#if NEED_SINT32
    ,{ 'l', put_sint32 }
#endif
#if NEED_SFRAC32
    ,{ 'Q', put_sfrac32 }
#endif
#if NEED_HEX8
    ,{ 'Y', put_hex8 }
#endif
    ,{ '\0', 0 }
};


#define my_va_arg_2(ARGP)       \
({                              \
    uint16_t arg;               \
    __asm ("ld %A0, %a1+  \n\t" \
         "ld %B0, %a1+  \n\t"   \
        : "=&r" (arg),          \
          "+e"  (ARGP));        \
    arg;                        \
})

#define my_va_arg_4(ARGP)       \
({                              \
    uint32_t arg;               \
    __asm ("ld %A0, %a1+  \n\t" \
         "ld %B0, %a1+  \n\t"   \
         "ld %C0, %a1+  \n\t"   \
         "ld %D0, %a1+  \n\t"   \
        : "=&r" (arg),          \
          "+e"  (ARGP));        \
    arg;                        \
})

#define my_va_arg_2_4(ARGP)     \
({                              \
    uint32_t arg;               \
    __asm ("clr %A0       \n\t" \
         "clr %B0       \n\t"   \
         "ld %C0, %a1+  \n\t"   \
         "ld %D0, %a1+  \n\t"   \
        : "=&r" (arg),          \
          "+e"  (ARGP));        \
    arg;                        \
})

// Decode format string <fmt> and dispatch to respective putter

static void //__attribute__((noinline))
put_va (const char * fmt, va_list args)
{
    put_local = put_char;
    
    while (1)
    {
        char c = pgm_read_byte_inc (fmt);
        
        if ('\0' == c)
            return;

        if ('%' == c)
        {
            c = pgm_read_byte_inc (fmt);

            if ('%' == c)
            {
                put_char2 (c);
                continue;
            }
#if NEED_MEM            
            if ('M' == c)
            {
                put_mem();
                continue;
            }
#endif // MEM

#if HAVE_32BIT_FUNCS
            uint8_t _32_bit = 0;
#endif
            const struct putters * p = fn_map;

            while (1)
            {
                // read id
                const char id = pgm_read_byte_inc (p); 
                // read func
                const putter_t pfunc = (putter_t) pgm_read_word_inc (p);
                
                if (id == c)
                {
                    // id found: dispatch to putter func
                    putter_arg_t arg;
#if HAVE_32BIT_FUNCS
                    if (_32_bit)
                        arg = my_va_arg_4 (args);
                    else
                        arg = my_va_arg_2_4 (args);
#else
                    arg = my_va_arg_2 (args);
#endif // HAVE_32BIT_FUNCS

                    pfunc (arg);
                    break;
                }

#if HAVE_32BIT_FUNCS
                if ('c' == id)
                    _32_bit = 1;
#endif // HAVE_32BIT_FUNCS

                // Unknown id: stop output
                if (id == '\0')
                {
                    put_char2 ('?');
                    put_char2 (c);
                    return;
                }
            }

            continue;
        } // c == '%'

        if ('\n' == c)
        {
            HIDE (c);
            put_char2 ('\r');
        }
        put_char2 (c);
    }
}

void put_PSTR (const char * fmt, ...)
{
    va_list vlist;
    va_start (vlist, fmt);

    put_va (fmt, vlist);

    va_end (vlist);
}
